﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;

namespace SharpDX3D11Basic1
{
	class MainForm : SharpDX.Windows.RenderForm
	{
		public void Run()
		{
			this.初期化する();

			// メインループ。
			SharpDX.Windows.RenderLoop.Run( this, () => {
				this.D3Dデバイスが消失していれば再構築する();	// 再構築できないなら this.Exit = true で帰ってくる。
				if( this.Exit )
				{
					this.終了する();
					this.Close();
					return;
				}
				this.シーンを描画する();
				this.swapChain.Present( 0, SharpDX.DXGI.PresentFlags.None );
			} );
		}

		protected System.Drawing.Size initialFormSize = new System.Drawing.Size( 1024, 1024 );
		protected bool Exit;

		protected SharpDX.Direct3D11.Device d3dDevice = null;
		protected SharpDX.DXGI.SwapChain swapChain = null;
		protected SharpDX.Direct3D11.RenderTargetView renderView = null;
		protected SharpDX.Direct3D11.Texture2D depthBuffer = null;
		protected SharpDX.Direct3D11.DepthStencilView depthView = null;

		protected SharpDX.Matrix View;
		protected SharpDX.Matrix Projection;
		protected SharpDX.Matrix ViewProjection;

		protected void 初期化する()
		{
			this.Text = "SharpDX3D11DesktopBasic1";
			this.ClientSize = this.initialFormSize;

			this.デバイスリソースを作成する();

			// キー入力ハンドラを登録する。
			this.KeyDown += ( target, arg ) => {
				// Esc → アプリ終了。
				if( arg.KeyCode == System.Windows.Forms.Keys.Escape )
					this.Exit = true;
			};
		}
		protected void 終了する()
		{
			this.デバイスリソースを解放する();
		}
		protected void デバイスリソースを作成する()
		{
			// スワップチェーン desc
			var swapChainDesc = new SharpDX.DXGI.SwapChainDescription() {
				BufferCount = 2,
				ModeDescription = new SharpDX.DXGI.ModeDescription(
					this.ClientSize.Width,
					this.ClientSize.Height,
					new SharpDX.DXGI.Rational( 60, 1 ),
					SharpDX.DXGI.Format.R8G8B8A8_UNorm ),
				IsWindowed = true,
				OutputHandle = this.Handle,
				SampleDescription = new SharpDX.DXGI.SampleDescription( 1, 0 ),
				SwapEffect = SharpDX.DXGI.SwapEffect.Discard,
				Usage = SharpDX.DXGI.Usage.RenderTargetOutput,
			};

			// 機能レベル
			var featureLevels = new SharpDX.Direct3D.FeatureLevel[] {
				SharpDX.Direct3D.FeatureLevel.Level_11_0,
				SharpDX.Direct3D.FeatureLevel.Level_10_1,
				SharpDX.Direct3D.FeatureLevel.Level_10_0,
			};

			// デバイスとスワップチェーンを作成する。
			SharpDX.Direct3D11.Device.CreateWithSwapChain(
				SharpDX.Direct3D.DriverType.Hardware,
				SharpDX.Direct3D11.DeviceCreationFlags.None,
				featureLevels,
				swapChainDesc,
				out this.d3dDevice,
				out this.swapChain );

			Trace.WriteLine( "D3Dデバイスとスワップチェーンを生成しました。" );
			Trace.WriteLine( $"機能レベル: {this.d3dDevice.FeatureLevel.ToString()}" );

			// すべての Windows イベントを無視する。具体的には PrintScreen と Alt+Enter 。
			using( var factory = this.swapChain.GetParent<SharpDX.DXGI.Factory>() )
			{
				factory.MakeWindowAssociation( this.Handle, SharpDX.DXGI.WindowAssociationFlags.IgnoreAll );
			}

			// バックバッファから RenderTargetView を作成する。
			using( var backBuffer = SharpDX.Direct3D11.Texture2D.FromSwapChain<SharpDX.Direct3D11.Texture2D>( this.swapChain, 0 ) )
			{
				this.renderView = new SharpDX.Direct3D11.RenderTargetView( this.d3dDevice, backBuffer );
			}

			// 深度バッファと深度ステンシルビューを作成する。
			this.depthBuffer = new SharpDX.Direct3D11.Texture2D(
				this.d3dDevice,
				new SharpDX.Direct3D11.Texture2DDescription() {
					Format = SharpDX.DXGI.Format.D32_Float_S8X24_UInt,
					ArraySize = 1,
					MipLevels = 1,
					Width = this.ClientSize.Width,
					Height = this.ClientSize.Height,
					SampleDescription = new SharpDX.DXGI.SampleDescription( 1, 0 ),
					Usage = SharpDX.Direct3D11.ResourceUsage.Default,
					BindFlags = SharpDX.Direct3D11.BindFlags.DepthStencil,
					CpuAccessFlags = SharpDX.Direct3D11.CpuAccessFlags.None,
					OptionFlags = SharpDX.Direct3D11.ResourceOptionFlags.None,
				} );
			this.depthView = new SharpDX.Direct3D11.DepthStencilView(
				this.d3dDevice,
				this.depthBuffer );

			// 変換行列を準備する。
			const float viewZ = 5.0f;
			this.View = SharpDX.Matrix.LookAtLH( new SharpDX.Vector3( 0, 0, -viewZ ), new SharpDX.Vector3( 0, 0, 0 ), SharpDX.Vector3.UnitY );
			this.Projection = SharpDX.Matrix.PerspectiveFovLH( (float) Math.PI / 4.0f, this.ClientSize.Width / (float) this.ClientSize.Height, 0.1f, 100.0f );
			this.ViewProjection = SharpDX.Matrix.Multiply( this.View, this.Projection );
		}
		protected void デバイスリソースを解放する()
		{
			this.d3dDevice?.ImmediateContext?.ClearState();
			this.swapChain?.SetFullscreenState( fullscreen: false, targetRef: null );   // ウィンドウモードにする。

			this.depthView?.Dispose();
			this.depthBuffer?.Dispose();
			this.renderView?.Dispose();
			this.swapChain?.Dispose();
			this.d3dDevice?.Dispose();
		}
		protected void シーンを描画する()
		{
			// レンダーターゲットビューを黒でクリアする。
			this.d3dDevice.ImmediateContext.ClearRenderTargetView(
				this.renderView,
				SharpDX.Color4.Black );

			// 深度ステンシルビューをクリアする。（ここでは深度バッファのみクリア。）
			this.d3dDevice.ImmediateContext.ClearDepthStencilView( 
				this.depthView, 
				SharpDX.Direct3D11.DepthStencilClearFlags.Depth,
				depth: 1.0f,
				stencil: 0 );

			// todo: シーンを描画する。
		}
		protected void D3Dデバイスが消失していれば再構築する()
		{
			var 削除理由 = this.d3dDevice.DeviceRemovedReason;

			if( 削除理由.Failure )
			{
				var エラー詳細 = new[] {
					new { Code = SharpDX.DXGI.ResultCode.DeviceHung.Code, Info = SharpDX.DXGI.ResultCode.DeviceHung.ApiCode, Rebuild = true },
					new { Code = SharpDX.DXGI.ResultCode.DeviceReset.Code, Info = SharpDX.DXGI.ResultCode.DeviceReset.ApiCode, Rebuild = true },
					new { Code = SharpDX.DXGI.ResultCode.DeviceRemoved.Code, Info = SharpDX.DXGI.ResultCode.DeviceRemoved.ApiCode, Rebuild = false },
					new { Code = SharpDX.DXGI.ResultCode.DriverInternalError.Code, Info = SharpDX.DXGI.ResultCode.DriverInternalError.ApiCode, Rebuild = false },
					new { Code = SharpDX.DXGI.ResultCode.InvalidCall.Code, Info = SharpDX.DXGI.ResultCode.InvalidCall.ApiCode, Rebuild = false },
				}.First( ( エラー ) => エラー.Code == 削除理由.Code );  // 見つからないなら System.InvalidOperationException 。

				Trace.WriteLine( $"D3Dデバイスが消失しました: {エラー詳細.Info}" );

				if( エラー詳細.Rebuild )
				{
					this.デバイスリソースを解放する();
					this.デバイスリソースを作成する();
				}
				else
				{
					this.Exit = true;
				}
			}
		}
	}
}
